# coding: utf-8

from functools import partial

from aiohttp import ClientSession
from thrift.protocol.TBinaryProtocol import TBinaryProtocol
from thrift.Thrift import TApplicationException

from .platform.thrift import (
    serialize, deserialize, convert_args_to_kwargs, get_call_result, ThriftService
)


class AsyncNexusClient(object):

    def __init__(self, services: list, address: tuple, timeout=None, protocol_cls=TBinaryProtocol):
        """Initialize AsyncNexusClient

        :param services: A list of thrift service module.
        :param address: A (host, port) tuple.
        :param timeout: Timeout seconds.
        :param protocol_cls: Thrift protocol class, default is `TBinaryProtocol`.
        """
        self.address = address
        self.protocol_cls = protocol_cls
        self.timeout = timeout
        self.httpclient = ClientSession()
        self.services = {}
        for service_module in services:
            service = ThriftService(service_module)
            self.services[service.name] = _AsyncNexusClientService(service, self)

    def __getattr__(self, item: str):
        if item not in self.services:
            raise AttributeError(item)
        return self.services[item]

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    def close(self):
        return self.httpclient.close()


class _AsyncNexusClientService(object):

    def __init__(self, service: ThriftService, client: AsyncNexusClient):
        self.service = service
        self.client = client

    def __getattr__(self, item: str):
        if not self.service.has_rpc(item):
            raise AttributeError('service {!r} has no api {!r}'.format(self.service.name, item))
        return partial(self._call, item)

    async def _call(self, rpc_name: str, *args, **kwargs):
        service_name = self.service.name
        url = 'http://{}:{}/{}/{}'.format(*self.client.address, service_name, rpc_name)
        rpc_args, rpc_result = self.service.get_rpc_args_and_result_object(rpc_name)

        for k, v in convert_args_to_kwargs(rpc_args, *args, **kwargs).items():
            setattr(rpc_args, k, v)

        data = serialize(rpc_args, self.client.protocol_cls)

        async with self.client.httpclient.post(url, data=data, timeout=self.client.timeout) as resp:
            if resp.status == 404:
                raise TApplicationException(TApplicationException.UNKNOWN_METHOD)
            if resp.status == 500:
                raise TApplicationException(TApplicationException.INTERNAL_ERROR)
            assert resp.status < 400, "Unknown HTTP Error"
            deserialize(rpc_result, await resp.read(), self.client.protocol_cls)

        return get_call_result(rpc_result)



# coding: utf-8

import logging
from typing import Dict
from inspect import iscoroutine

from thrift.protocol.TBinaryProtocol import TBinaryProtocol
from aiohttp.web import Application, Request, Response, run_app
from aiohttp.web_exceptions import HTTPNotFound, HTTPInternalServerError

from .platform.thrift import serialize, deserialize, get_call_args, ThriftService

logger = logging.getLogger(__name__)


class AsyncNexusServer(object):

    def __init__(self, services_map: list, address: tuple, protocol_cls=TBinaryProtocol):
        """Initialize AsyncNexusServer

        :param services_map: A list of (thrift_service, api_handler) two-tuples.
        :param address: A (host, port) tuple.
        :param protocol_cls: Thrift protocol class, default is `TBinaryProtocol`.
        """
        self.services_map: Dict[str, ThriftService] = {}
        for service_module, handler in services_map:
            service = ThriftService(service_module, handler)
            self.services_map[service.name] = service
        self.address = address
        self.protocol_cls = protocol_cls
        self._app = Application()
        self._app.router.add_post('/{service}/{rpc}', self._handle_request)

    def _has_service(self, service_name: str) -> bool:
        return service_name in self.services_map

    @staticmethod
    async def _process(rpc_impl, call_args):
        ret = rpc_impl(*call_args)
        if iscoroutine(ret):
            return await ret
        return ret

    async def _handle_request(self, request: Request):
        service_name = request.match_info['service']
        rpc_name = request.match_info['rpc']

        if not self._has_service(service_name):
            raise HTTPNotFound(body=b'')

        service = self.services_map[service_name]

        if not service.has_rpc(rpc_name):
            raise HTTPNotFound(body=b'')

        rpc_impl = getattr(service.handler, rpc_name)
        rpc_args, rpc_result = service.get_rpc_args_and_result_object(rpc_name)

        deserialize(rpc_args, await request.read(), self.protocol_cls)

        call_args = get_call_args(rpc_args)

        try:
            rpc_result.success = await self._process(rpc_impl, call_args)
        except Exception as e:
            for _, _, exc_name, exc_type_info, *_ in rpc_result.thrift_spec:
                if exc_name == 'success':
                    continue
                exc_class, *_ = exc_type_info
                if isinstance(e, exc_class):
                    setattr(rpc_result, exc_name, e)
                    break
            else:
                logger.exception('NexusServiceError')
                raise HTTPInternalServerError(body=b'') from e

        return Response(body=serialize(rpc_result, self.protocol_cls))

    def run(self):
        run_app(self._app, host=self.address[0], port=self.address[1])



# coding: utf-8

from typing import Dict

from thrift.protocol.TBinaryProtocol import TBinaryProtocol
from thrift.TSerialization import (
    serialize as thrift_serialize,
    deserialize as thrift_deserialize,
)


class TProtocolFactory(object):

    def __init__(self, protocol_cls):
        """Initialize TProtocolFactory

        :param protocol_cls: Thrift protocol class.
        """
        self.protocol_cls = protocol_cls

    def get_protocol(self, trans):
        return self.protocol_cls(trans)

    getProtocol = get_protocol


class ThriftService(object):

    def __init__(self, service_module, handler=None):
        self.name = service_module.__name__.split('.')[-1]
        self.service_module = service_module
        self.handler = handler
        self._service_rpc_map: Dict[str, tuple] = {}
        for rpc_name in service_module.Iface.__dict__:
            rpc_args = getattr(service_module, rpc_name + '_args', None)
            rpc_result = getattr(service_module, rpc_name + '_result', None)
            if rpc_args is None or rpc_result is None:
                continue
            self._service_rpc_map[rpc_name] = (rpc_args, rpc_result)

    def has_rpc(self, rpc_name: str) -> bool:
        return rpc_name in self._service_rpc_map

    def get_rpc_args_and_result_object(self, rpc_name: str) -> tuple:
        rpc_args, rpc_result = self._service_rpc_map[rpc_name]
        return rpc_args(), rpc_result()


def serialize(thrift_object, protocol_cls=TBinaryProtocol) -> bytes:
    return thrift_serialize(thrift_object, TProtocolFactory(protocol_cls))


def deserialize(thrift_object, buf: bytes, protocol_cls=TBinaryProtocol):
    return thrift_deserialize(thrift_object, buf, TProtocolFactory(protocol_cls))


def convert_args_to_kwargs(thrift_args_object, *args, **kwargs):
    arg_names = [_[2] for _ in thrift_args_object.thrift_spec if _]
    return {**dict(zip(arg_names, args)), **kwargs}


def get_call_args(thrift_args_object) -> list:
    """From a thrift args object get call positional arguments"""
    arg_names = [_[2] for _ in thrift_args_object.thrift_spec if _]
    return [getattr(thrift_args_object, arg_name) for arg_name in arg_names]


def get_call_result(thrift_result_object):
    if getattr(thrift_result_object, 'success', None) is not None:
        return thrift_result_object.success
    for k, v in thrift_result_object.__dict__.items():
        if k != 'success' and v:
            raise v
    return None



#!/usr/bin/env python
# coding: utf-8

from setuptools import setup

import nexus


setup(
    name='nexus-rpc',
    version=nexus.__version__,
    packages=['nexus', 'nexus.platform'],
    author='JZQT',
    author_email='561484726@qq.com',
    url="https://github.com/JZQT/nexus-rpc",
    description="Nexus is a http-based rpc framework.",
    license='MIT',
    install_requires=['aiohttp', 'thrift'],
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ]
)
